#include <linux/vmalloc.h>
#include <linux/list.h>
#include "ldfs_lcache.h"
#include "ldfs_common.h"
#include "ldfs_store.h"
#include "ldfs_slist.h"

static inline ext4_fsblk_t ldfs_lcache_blocknr_get(struct super_block * sb,
		struct ldfs_lcache_group * cache_group, struct ldfs_lcache_entry * entry) {
	return cache_group->g_groupnr * LDFS_STORE_ENTRIES(sb) + (entry - cache_group->g_entry);
}

static void ldfs_lcache_group_wait(struct ldfs_lcache_group * cache_group) {
	mutex_lock(&cache_group->g_load_mutex);
	atomic_inc(&cache_group->g_count);
	mutex_unlock(&cache_group->g_load_mutex);
}

int ldfs_lcache_insert_fp(struct super_block * sb, struct ldfs_lcache_group * cache_group,
		ext4_fsblk_t blocknr, __u8 fp[], __u32 cnt) {
	struct ldfs_lcache * cache = EXT4_SB(sb)->s_lcache;
	ext4_grpblk_t offset = blocknr % LDFS_STORE_ENTRIES(sb);
	struct ldfs_lcache_entry * entry = &cache_group->g_entry[offset];
	int p = LDFS_LCACHE_ENTRY_POS(fp);

	if (cnt) {
		memcpy(entry->e_fp, fp, LDFS_FP_LEN_MAX);
		spin_lock(&cache->l_entry_lock[LDFS_LCACHE_LOCK(fp)]);
		slist_add(&entry->e_head, &cache_group->g_entry_head[p]);
		spin_unlock(&cache->l_entry_lock[LDFS_LCACHE_LOCK(fp)]);
	} else {
		entry->e_head.next = NULL;
		//slist_head_init(&entry->e_head);
	}
	return 0;
}

int ldfs_lcache_group_load(struct super_block * sb, struct ldfs_lcache_group * cache_group) {
	struct ldfs_lcache * cache = EXT4_SB(sb)->s_lcache;
	ext4_fsblk_t start_blocknr = cache_group->g_groupnr * LDFS_STORE_ENTRIES(sb);
	int blocks = LDFS_LCACHE_DIV_WIDTH(sb);
	int my_div;

	while ((my_div = atomic_inc_return(&cache_group->g_load_div) - 1) < LDFS_LCACHE_DIV_CNT) {
		ext4_fsblk_t blocknr = start_blocknr + my_div * blocks;
		ldfs_store_fetch(sb, blocknr, blocks, ldfs_lcache_insert_fp, cache_group);
	}

	spin_lock(&cache->l_group_lock);
	if (atomic_dec_and_test(&cache_group->g_count)) {
		list_del(&cache_group->g_head);
		list_add(&cache_group->g_head, &cache->l_using_head);
		mutex_unlock(&cache_group->g_load_mutex);
	}
	spin_unlock(&cache->l_group_lock);
	return 0;
}

struct ldfs_lcache_group * ldfs_lcache_group_get(struct ldfs_lcache * cache, ext4_fsblk_t groupnr) {
	struct ldfs_lcache_group * cache_group, * n;
	int i;

	spin_lock(&cache->l_group_lock);
	/* Search in using groups, high probability as writes have spatial locality */
	list_for_each_entry(cache_group, &cache->l_using_head, g_head) {
		if (cache_group->g_groupnr == groupnr) {
			atomic_inc(&cache_group->g_count);
			spin_unlock(&cache->l_group_lock);
			return cache_group;
		}
	}

	/* Search in groups which is accessed before, using temporal locality */
	list_for_each_entry_safe(cache_group, n, &cache->l_wait_head, g_head) {
		if (cache_group->g_groupnr == groupnr && cache_group->valid) {
			list_del(&cache_group->g_head);
			list_add(&cache_group->g_head, &cache->l_using_head);
			atomic_inc(&cache_group->g_count);
			spin_unlock(&cache->l_group_lock);
			return cache_group;
		}
	}

	/* Check if the group is reloading */
	list_for_each_entry(cache_group, &cache->l_reload_head, g_head) {
		if (cache_group->g_groupnr == groupnr) {
#ifdef LDFS_LCACHE_SMP
			atomic_inc(&cache_group->g_count);
			spin_unlock(&cache->l_group_lock);
			ldfs_lcache_group_load(cache->l_sb, cache_group);
#else
			spin_unlock(&cache->l_group_lock);
#endif
			ldfs_lcache_group_wait(cache_group);
			return cache_group;
		}
	}

	/* Load group from disk */
	if (!list_empty(&cache->l_free_head)) {
		cache_group = list_first_entry(&cache->l_free_head, struct ldfs_lcache_group, g_head);
	} else {
		BUG_ON(list_empty(&cache->l_wait_head));
		cache_group = list_entry(cache->l_wait_head.prev, struct ldfs_lcache_group, g_head);
	}
	list_del(&cache_group->g_head);
	list_add(&cache_group->g_head, &cache->l_reload_head);
	cache_group->g_groupnr = groupnr;
	atomic_set(&cache_group->g_count, 1);
	atomic_set(&cache_group->g_load_div, 0);
	cache_group->valid = 1;
	mutex_lock(&cache_group->g_load_mutex);

	/* For performance reason, this initialization should not be here */
	for (i = 0; i < LDFS_LCACHE_SEARCH_HEAD; i++) {
		slist_head_init(&cache_group->g_entry_head[i]);
	}
	spin_unlock(&cache->l_group_lock);

	ldfs_lcache_group_load(cache->l_sb, cache_group);
	ldfs_lcache_group_wait(cache_group);
	return cache_group;
}

void ldfs_lcache_group_put(struct ldfs_lcache * cache, struct ldfs_lcache_group * cache_group) {
	spin_lock(&cache->l_group_lock);
	if (atomic_dec_and_test(&cache_group->g_count)) {
		list_del(&cache_group->g_head);
		list_add(&cache_group->g_head, &cache->l_wait_head);
	}
	spin_unlock(&cache->l_group_lock);
}

int ldfs_lcache_search_group(struct super_block * sb, __u8 fp[], ext4_fsblk_t groupnr,
		ext4_fsblk_t * blocknr) {
	struct ldfs_lcache * cache = EXT4_SB(sb)->s_lcache;
	struct ldfs_lcache_group * cache_group = ldfs_lcache_group_get(cache, groupnr);
	struct ldfs_lcache_entry * entry;
	int p = LDFS_LCACHE_ENTRY_POS(fp);
	struct slist_head * head;

	spin_lock(&cache->l_entry_lock[LDFS_LCACHE_LOCK(fp)]);
	slist_for_each(head, &cache_group->g_entry_head[p]) {
		entry = slist_entry(head, struct ldfs_lcache_entry, e_head);
		if (ldfs_fp_equal(sb, fp, entry->e_fp)) {
			*blocknr = ldfs_lcache_blocknr_get(sb, cache_group, entry);
			spin_unlock(&cache->l_entry_lock[LDFS_LCACHE_LOCK(fp)]);
			ldfs_lcache_group_put(cache, cache_group);
			return 0;
		}
	}
	spin_unlock(&cache->l_entry_lock[LDFS_LCACHE_LOCK(fp)]);
	ldfs_lcache_group_put(cache, cache_group);
	return -1;
}

int ldfs_lcache_invalidate(struct super_block * sb, ext4_fsblk_t blocknr) {
	struct ldfs_lcache * cache = EXT4_SB(sb)->s_lcache;
	ext4_fsblk_t groupnr = blocknr / LDFS_STORE_ENTRIES(sb);
	struct ldfs_lcache_group * cache_group;

	spin_lock(&cache->l_group_lock);
	/* Search in using groups, high probability as writes have spatial locality */
	list_for_each_entry(cache_group, &cache->l_using_head, g_head) {
		if (cache_group->g_groupnr == groupnr) {
			cache_group->valid = 0;
			spin_unlock(&cache->l_group_lock);
			return 0;
		}
	}
	spin_unlock(&cache->l_group_lock);
	return 0;
}

int ldfs_lcache_init(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	struct ldfs_lcache * cache;
	struct ldfs_lcache_group * cache_group;
	struct list_head * head;
	int i, j;

	/* Create the cache structure */
	cache = kmalloc(sizeof(struct ldfs_lcache), GFP_KERNEL);
	sbi->s_lcache = cache;
	cache->l_sb = sb;
	INIT_LIST_HEAD(&cache->l_using_head);
	INIT_LIST_HEAD(&cache->l_wait_head);
	INIT_LIST_HEAD(&cache->l_reload_head);
	INIT_LIST_HEAD(&cache->l_free_head);
	spin_lock_init(&cache->l_group_lock);

	for (i = 0; i < LDFS_LCACHE_LOCK_CNT; i++) {
		spin_lock_init(&cache->l_entry_lock[i]);
	}

	/* Create group structure */
	for (i = 0; i < LDFS_LCACHE_GROUP_COUNT; i++) {
		cache_group = kmalloc(sizeof(struct ldfs_lcache_group), GFP_KERNEL);
		INIT_LIST_HEAD(&cache_group->g_head);
		mutex_init(&cache_group->g_load_mutex);
		for (j = 0; j < LDFS_LCACHE_SEARCH_HEAD; j++) {
			slist_head_init(&cache_group->g_entry_head[j]);
		}
		list_add(&cache_group->g_head, &cache->l_free_head);
	}

	/* Create entry structure */list_for_each(head, &cache->l_free_head) {
		cache_group = list_entry(head, struct ldfs_lcache_group, g_head);
		cache_group->g_entry = vmalloc(LDFS_STORE_ENTRIES(sb) * sizeof(struct ldfs_lcache_entry));
	}
	return 0;
}

int ldfs_lcache_exit(struct super_block * sb) {
	struct ext4_sb_info * sbi = EXT4_SB(sb);
	struct ldfs_lcache * cache = sbi->s_lcache;
	struct ldfs_lcache_group * cache_group;
	struct list_head * head, *n;

	/* Remove entry and group structure */list_for_each_safe(head, n, &cache->l_using_head) {
		cache_group = list_entry(head, struct ldfs_lcache_group, g_head);
		vfree(cache_group->g_entry);
		list_del(head);
		kfree(cache_group);
	}

	list_for_each_safe(head, n, &cache->l_wait_head) {
		cache_group = list_entry(head, struct ldfs_lcache_group, g_head);
		vfree(cache_group->g_entry);
		list_del(head);
		kfree(cache_group);
	}

	list_for_each_safe(head, n, &cache->l_reload_head) {
		cache_group = list_entry(head, struct ldfs_lcache_group, g_head);
		vfree(cache_group->g_entry);
		list_del(head);
		kfree(cache_group);
	}

	list_for_each_safe(head, n, &cache->l_free_head) {
		cache_group = list_entry(head, struct ldfs_lcache_group, g_head);
		vfree(cache_group->g_entry);
		list_del(head);
		kfree(cache_group);
	}

	kfree(cache);
	return 0;
}
